# -*- coding: utf-8; -*-

from decimal import Decimal
from collections import deque
from copy import copy
from io import BytesIO

import xml.sax
import xml.sax.handler

import datetime
import re
import iso8601


def format_open_tag(name, attrs):
    attrs = [u'%s="%s"' % attr for attr in attrs.items()]
    return u"<%s %s>" % (name, ' '.join(attrs))


class ElementHandler(object):
    __slots__ = ('value', )

    def __init__(self):
        self.value = copy(self.default)

    def on_nested_start(self, name, attrs):
        raise NotImplementedError(
            "Elements with type=%s are not expected to have nested elements" % self.type_name)

    def on_nested_end(self, name, handler):
        pass

    def on_content(self, content):
        if not content.strip():
            return
        raise NotImplementedError(
            "Elements with type=%s are not expected to have a content" % self.type_name)


class NoTypeHandler(ElementHandler):
    __slots__ = ('_string_parts', '_dict')

    type_name = None
    default = None

    wspace_re = re.compile('\s+')

    def __init__(self):
        self._string_parts = []
        self._dict = {}

    def on_nested_start(self, name, attrs):
        if self._string_parts:
            self._string_parts.append(format_open_tag(name, attrs))
            return MixedContentHandler()
        return handler_for(attrs)

    def on_nested_end(self, name, handler):
        if self._string_parts:
            self._string_parts.append(handler.value)
            self._string_parts.append("</%s>" % name)
            return
        self._dict[name] = handler.value

    def on_content(self, content):
        if self._dict:
            # Hack: drop content if we're already got nested
            # elements. This could lead to a bug, but since
            # InSales doesn't enclose nested HTML with CDATA
            # there is no easy way to recover string data as-is
            # in such case
            return
        if not self._string_parts and not content.strip():
            # skip insignificant leading whitespace
            return
        self._string_parts.append(content)

    @property
    def value(self):
        if self._dict:
            return self._dict
        if self._string_parts:
            val = u''.join(self._string_parts)
            val = self.wspace_re.sub(u' ', val)
            val = val.strip()
            return val
        return self.default


class MixedContentHandler(ElementHandler):
    default = ''

    def on_nested_start(self, name, attrs):
        self.value += format_open_tag(name, attrs)
        return MixedContentHandler()

    def on_nested_end(self, name, handler):
        self.value += handler.value
        self.value += "</%s>" % name

    def on_content(self, content):
        content = content.strip()
        self.value += content


class NilHandler(ElementHandler):
    default = None

    def on_nested_start(self, name, attrs):
        raise NotImplementedError(
            "Elements with nil=true are not expected to have nested elements")

    def on_nested_end(self, name, handler):
        pass

    def on_content(self, content):
        if not content.strip():
            return
        raise NotImplementedError(
            "Elements with nil=true are not expected to have a content")


class ArrayHandler(ElementHandler):
    type_name = 'array'
    default = []

    def on_nested_start(self, name, attrs):
        return handler_for(attrs)

    def on_nested_end(self, name, handler):
        self.value.append(handler.value)


class IntegerHandler(ElementHandler):
    type_name = 'integer'
    default = 0

    def on_content(self, content):
        self.value = int(content.strip())


class DecimalHandler(ElementHandler):
    type_name = 'decimal'
    default = Decimal(0)

    def on_content(self, content):
        self.value = Decimal(content.strip())


class BooleanHandler(ElementHandler):
    type_name = 'boolean'
    default = False

    def on_content(self, content):
        self.value = (content.strip() == 'true')


class DateHandler(ElementHandler):
    type_name = 'date'
    default = None

    def on_content(self, content):
        self.value = datetime.datetime.strptime(content.strip(), "%Y-%m-%d")


class TimestampHandler(ElementHandler):
    type_name = 'timestamp'
    default = None
    date_re = re.compile(r"\s+\+(\d\d)(\d\d)$")

    def on_content(self, content):
        # convert 2010-08-16 18:39:58 +0400
        # to      2010-08-16 18:39:58+04:00
        string = self.date_re.sub(r"+\1:\2", content.strip())
        self.value = iso8601.parse_date(string)

class DateTimeHandler(ElementHandler):
    type_name = 'dateTime'
    default = None

    def on_content(self, content):
        self.value = iso8601.parse_date(content.strip())


all_handlers = [
    NoTypeHandler,
    ArrayHandler,
    IntegerHandler,
    DecimalHandler,
    BooleanHandler,
    DateHandler,
    DateTimeHandler,
    TimestampHandler,
]

type2handler = dict((h.type_name, h) for h in all_handlers)

def handler_for(attrs):
    nil = (attrs.get('nil') == 'true')
    if nil:
        return NilHandler()
    type_name = attrs.get('type')
    return type2handler.get(type_name, NoTypeHandler)()


class XmlProcessor(xml.sax.handler.ContentHandler):
    def __init__(self):
        xml.sax.handler.ContentHandler.__init__(self)
        self._handler_stack = deque([NoTypeHandler()])

    def startElement(self, name, attrs):
        head = self._handler_stack[-1]
        new_head = head.on_nested_start(name, attrs)
        self._handler_stack.append(new_head)

    def endElement(self, name):
        h = self._handler_stack.pop()
        self._handler_stack[-1].on_nested_end(name, h)

    def characters(self, content):
        self._handler_stack[-1].on_content(content)

    def data(self):
        top_dict = self._handler_stack[0].value
        if top_dict:
            return list(top_dict.values())[0]


def parse(xml_string):
    processor = XmlProcessor()
    io = BytesIO(xml_string)

    parser = xml.sax.make_parser()
    parser.setContentHandler(processor)
    for line in io:
        parser.feed(line)

    return processor.data()



# -*- coding: utf-8; -*-

import datetime
import time
import socket

from base64 import b64encode

try:
    # Python 3
    from urllib import parse as urlparse
    from urllib.parse import urlencode
    from http.client import HTTPConnection, HTTPException
except ImportError:
    # Python 2
    import urlparse
    from httplib import HTTPConnection, HTTPException
    from urllib import urlencode


class ApiError(Exception):
    def __init__(self, msg, code=None):
        super(ApiError, self).__init__(msg)
        self.code = code


class Connection(object):
    def __init__(self, account, api_key, password,
                 retry_on_503=False, retry_on_socket_error=False,
                 retry_timeout=1, response_timeout=10):
        self.account = account
        self.api_key = api_key
        self.password = password
        self.retry_on_503 = retry_on_503
        self.retry_on_socket_error = retry_on_socket_error
        self.retry_timeout = retry_timeout
        self.response_timeout = response_timeout

    def request(self, method, endpoint, qargs={}, data=None):
        path = self.format_path(endpoint, qargs)
        conn = HTTPConnection('%s.myinsales.ru:80' % self.account, timeout=self.response_timeout)
        auth = b64encode(u"{0}:{1}".format(self.api_key, self.password).encode('utf-8')).decode('utf-8')
        headers = {
            'Authorization': 'Basic {0}'.format(auth),
            'Content-Type': 'application/xml'
        }

        done = False
        while not done:
            try:
                conn.request(method, path, headers=headers, body=data)
                resp = conn.getresponse()
                body = resp.read()
            except (socket.gaierror, socket.timeout, HTTPException):
                if self.retry_on_socket_error:
                    time.sleep(self.retry_timeout)
                    continue
                else:
                    raise

            if resp.status == 503 and self.retry_on_503:
                time.sleep(self.retry_timeout)
            else:
                done = True

        if 200 <= resp.status < 300:
            return body
        else:
            raise ApiError("%s request to %s returned: %s\n%s" %
                           (method, path, resp.status, body), resp.status)

    def format_path(self, endpoint, qargs):
        for key, val in qargs.items():
            if isinstance(val, datetime.datetime):
                qargs[key] = val.replace(microsecond=0).isoformat()

        url_parts = list(urlparse.urlparse(endpoint))
        query = dict(urlparse.parse_qsl(url_parts[4]))
        query.update(qargs)
        url_parts[4] = urlencode(query)

        return urlparse.urlunparse(url_parts)

    def get(self, path, qargs):
        return self.request('GET', path, qargs=qargs)

    def put(self, path, data):
        return self.request('PUT', path, data=data)

    def post(self, path, data):
        return self.request('POST', path, data=data)

    def delete(self, path):
        return self.request('DELETE', path)



# -*- coding: utf-8; -*-

from insales.parsing import parse
from insales.composing import compose
from insales.connection import Connection


class InSalesApi(object):

    arrays = {
        'variants'              : 'variant',
        'variants-attributes'   : 'variant',
        'options'               : 'option',
        'option-names'          : 'option-name',
        'option-values'         : 'option-value',
        'orders'                : 'order',
        'order-lines-attributes': 'order-line-attributes',
        'images'                : 'image',
    }

    @classmethod
    def from_credentials(cls, account, api_key, password, **kwargs):
        return cls(Connection(account, api_key, password, **kwargs))

    def __init__(self, connection):
        self.connection = connection

    #========================================================================
    # Заказы
    #========================================================================
    def get_orders(self, per_page=25, page=1):
        return self._get('/admin/orders.xml', {'per_page': per_page, 'page': page}) or []

    def get_order(self, order_id):
        return self._get('/admin/orders/%s.xml' % order_id)

    def update_order(self, order_id, order_data):
        return self._update('/admin/orders/%s.xml' % order_id, order_data, root='order')

    def delete_order(self, order_id):
        self._delete('/admin/orders/%s.xml' % order_id)

    def create_order(self, order_data):
        return self._add('/admin/orders.xml', order_data, root='order')

    def get_order_delivery_variants(self, order_data):
        return self._post('/admin/orders/delivery_variants.xml',
                          order_data, root='order')

    def get_order_payment_gateways(self, order_data):
        return self._post('/admin/orders/payment_gateways.xml',
                          order_data, root='order')

    #========================================================================
    # Поля заказов
    #========================================================================
    def get_orders_fields(self):
        return self._get('/admin/orders/fields.xml')

    #========================================================================
    # Категории на складе
    #========================================================================
    def get_categories(self):
        return self._get('/admin/categories.xml') or []

    def get_category(self, category_id):
        return self._get('/admin/categories/%s.xml' % category_id)

    def add_category(self, category_data):
        return self._add('/admin/categories.xml', category_data, root='category')

    def update_category(self, category_id, category_data):
        return self._update('/admin/categories/%s.xml' % category_id,
                            category_data, root='category')

    def delete_category(self, category_id):
        self._delete('/admin/categories/%s.xml' % category_id)

    #========================================================================
    # Категории на сайте
    #========================================================================
    def get_collections(self):
        return self._get('/admin/collections.xml') or []

    def get_collection(self, collection_id):
        return self._get('/admin/collections/%s.xml' % collection_id)

    def add_collection(self, collection_data):
        return self._add('/admin/collections.xml', collection_data, root='collection')

    def update_collection(self, collection_id, collection_data):
        return self._update('/admin/collections/%s.xml' % collection_id,
                            collection_data, root='collection')

    def delete_collection(self, collection_id):
        self._delete('/admin/collections/%s.xml' % collection_id)
    
    #========================================================================
    # Свойства товаров
    #========================================================================
    def get_option_names(self):
        return self._get('/admin/option_names.xml') or []

    def get_option_name(self, option_name_id):
        return self._get('/admin/option_names/%s.xml' % option_name_id)

    def add_option_name(self, option_name):
        return self._add('/admin/option_names.xml', option_name, root='option-name')

    def update_option_name(self, option_name_id, option_name_data):
        return self._update('/admin/option_names/%s.xml' % option_name_id,
                            option_name_data, root='option-name')

    def delete_option_name(self, option_name_id):
        self._delete('/admin/option_names/%s.xml' % option_name_id)

    #========================================================================
    # Значения свойств
    #========================================================================
    def get_option_values(self, option_name_id=None):
        if option_name_id:
            path = '/admin/option_names/%s/option_values.xml' % option_name_id
        else:
            path = '/admin/option_values.xml'
        return self._get(path) or []

    def get_option_value(self, option_name_id, option_value_id):
        return self._get('/admin/option_names/%s/option_values/%s.xml' %
                         (option_name_id, option_value_id))

    def add_option_value(self, option_name_id, option_value_data):
        return self._add('/admin/option_names/%s/option_values.xml' % option_name_id, 
                         option_value_data, root='option-value')

    def update_option_value(self, option_name_id, option_value_id, option_value_data):
        return self._update('/admin/option_names/%s/option_values/%s.xml' %
                            (option_name_id, option_value_id),
                            option_value_data, root='option_value')

    def delete_option_value(self, option_name_id, option_value_id):
        self._delete('/admin/option_names/%s/option_values/%s.xml' %
                     (option_name_id, option_value_id))

    #========================================================================
    # Товары
    #========================================================================
    def get_products(self, limit=50, page=1, updated_since=None):
        qargs = {
            'limit': limit,
            'page': page,
        }
        if updated_since:
            qargs['updated_since'] = updated_since
        return self._get('/admin/products.xml', qargs) or []

    def get_product(self, product_id):
        return self._get('/admin/products/%s.xml' % product_id)

    def add_product(self, product_data):
        return self._add('/admin/products.xml', product_data, root='product')

    def update_product(self, product_id, product_data):
        return self._update('/admin/products/%s.xml' % product_id, product_data,
                            root='product')

    def delete_product(self, product_id):
        return self._delete('/admin/products/%s.xml' % product_id)

    #========================================================================
    # Модификации товаров
    #========================================================================
    def get_product_variants(self, product_id):
        return self._get('/admin/products/%s/variants.xml' % product_id) or []

    def get_product_variant(self, product_id, variant_id):
        return self._get('/admin/products/%s/variants/%s.xml' % (product_id, variant_id))

    def add_product_variant(self, product_id, variant_data):
        return self._add('/admin/products/%s/variants.xml' % product_id,
                         variant_data, root='variant')

    def update_product_variant(self, product_id, variant_id, variant_data):
        return self._update('/admin/products/%s/variants/%s.xml' %
                            (product_id, variant_id),
                            variant_data, root='variant')

    def delete_product_variant(self, product_id, variant_id):
        self._delete('/admin/products/%s/variants/%s.xml' % (product_id, variant_id))

    #========================================================================
    # Изображения товара
    #========================================================================
    def get_product_images(self, product_id):
        return self._get('/admin/products/%s/images.xml' % product_id) or []

    def get_product_image(self, product_id, image_id):
        return self._get('/admin/products/%s/images/%s.xml' % (product_id, image_id))

    def add_product_image(self, product_id, image_data):
        return self._add('/admin/products/%s/images.xml' % product_id,
                         image_data, root='image')

    def update_product_image(self, product_id, image_id, image_data):
        return self._update('/admin/products/%s/images/%s.xml' % (product_id, image_id),
                            image_data, root='image')

    def delete_product_image(self, product_id, image_id):
        self._delete('/admin/products/%s/images/%s.xml' % (product_id, image_id))

    #========================================================================
    # Размещение товара
    #========================================================================
    def get_collects(self, product_id=None, collection_id=None, page=1):
        qargs = {}
        if product_id:
            qargs['product_id'] = product_id
        if collection_id:
            qargs['collection_id'] = collection_id

        qargs['page'] = page

        return self._get('/admin/collects.xml', qargs) or []

    def add_collect(self, collect_data):
        return self._add('/admin/collects.xml', collect_data, root='collect')

    def update_collect(self, collect_id, collect_data):
        return self._update('/admin/collects/%s.xml' % collect_id,
                            collect_data, root='collect')

    def delete_collect(self, collect_id):
        return self._delete('/admin/collects/%s.xml' % collect_id)

    #========================================================================
    # Аналогичные товары
    #========================================================================
    def get_similars(self, product_id):
        return self._get('/admin/products/%s/similars.xml' % product_id) or []

    def delete_similar(self, product_id, similar_product_id):
        return self._delete('/admin/products/%s/similars/%s.xml' %
                            (product_id, similar_product_id))

    #========================================================================
    # Сопутствующие товары
    #========================================================================
    def get_supplementaries(self, product_id):
        return self._get('/admin/products/%s/supplementaries.xml' % product_id) or []

    def delete_supplementary(self, product_id, supplementary_product_id):
        return self._delete('/admin/products/%s/supplementaries/%s.xml' %
                            (product_id, supplementary_product_id))

    #========================================================================
    # Веб-хуки
    #========================================================================
    def get_webhooks(self):
        return self._get('/admin/webhooks.xml') or []

    def get_webhook(self, webhook_id):
        return self._get('/admin/webhooks/%s.xml' % webhook_id)

    def add_webhook(self, webhook):
        return self._add('/admin/webhooks.xml', webhook, root='webhook')

    def update_webhook(self, webhook_id, webhook_data):
        return self._update('/admin/webhooks/%s.xml' % webhook_id,
                            webhook_data, root='webhook')

    def delete_webhook(self, webhook_id):
        self._delete('/admin/webhooks/%s.xml' % webhook_id)

    #========================================================================
    # Биллинг
    #========================================================================
    def get_recurring_application_charge(self):
        return self._get('/admin/recurring_application_charge.xml')

    def add_recurring_application_charge(self, recurring_application_charge_data):
        return self._add('/admin/recurring_application_charge.xml',
                         recurring_application_charge_data,
                         root='recurring-application-charge')

    def update_recurring_application_charge(self, recurring_application_charge_data):
        return self._update('/admin/recurring_application_charge.xml',
                            recurring_application_charge_data,
                            root='recurring-application-charge')

    #========================================================================
    def _get(self, endpoint, qargs={}):
        return self._req('get', endpoint, qargs)

    def _add(self, endpoint, data, root):
        xml = compose(data, root=root, arrays=self.arrays)
        return self._req('post', endpoint, xml)

    _post = _add

    def _update(self, endpoint, data, root):
        xml = compose(data, root=root, arrays=self.arrays)
        return self._req('put', endpoint, xml)

    def _delete(self, endpoint):
        return self._req('delete', endpoint)

    def _req(self, method, *args, **kwargs):
        response = getattr(self.connection, method)(*args, **kwargs)
        return parse(response)



# -*- coding: utf-8; -*-

import datetime
import collections
import xml.etree.ElementTree as et

from decimal import Decimal

# Python 2, 3 intercompatibility
try:
    basestring
except NameError:
    basestring = str

try:
    long
except NameError:
    long = int

def compose(data, root, arrays={}):
    root_e = compose_element(root, data, arrays)
    return et.tostring(root_e, 'utf-8')

def compose_element(key, value, arrays={}):
    e = et.Element(key)
    if isinstance(value, basestring):
        e.text = value
    elif isinstance(value, (int, long)):
        e.attrib['type'] = 'integer'
        e.text = str(value)
    elif isinstance(value, Decimal):
        e.attrib['type'] = 'decimal'
        e.text = str(value)
    elif isinstance(value, datetime.datetime):
        e.attrib['type'] = 'dateTime'
        e.text = value.replace(microsecond=0).isoformat()
    elif value is None:
        e.attrib['nil'] = 'true'
    elif isinstance(value, collections.Sequence):
        e.attrib['type'] = 'array'
        e_key = arrays[key]
        for x in value:
            e.append(compose_element(e_key, x, arrays))
    elif isinstance(value, collections.Mapping):
        for key, value in value.items():
            e.append(compose_element(key, value, arrays))
    else:
        raise TypeError("Value %r has unsupported type %s" % (value, type(value)))
    return e



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

install_requires = open("requirements.txt").read().split('\n')
readme_content = open("README.md").read()

setup(
    name='pyinsales',
    version='1.0.0',
    description='InSales e-commerce platform API bindings',
    long_description=readme_content,
    author='Victor Nakoryakov',
    author_email='nail.xx@gmail.com',
    license='MIT',
    keywords="insales API bindings",
    url='https://github.com/nailxx/pyinsales',
    packages=['insales'],
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Natural Language :: Russian",
        "Topic :: Utilities",
    ],
)
